/* Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * $Id: getaddrinfo.c,v 1.7 2004/01/03 05:28:31 musirc Exp $ */

#include "stdinc.h"
#include <stdarg.h>
#include "getinfo.h"

static const char in_addrany[] = { 0, 0, 0, 0 };
static const char in_loopback[] = { 127, 0, 0, 1 };
static const char in6_addrany[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
static const char in6_loopback[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1
};

static const struct afd {
	int a_af;
	int a_addrlen;
	int a_socklen;
	int a_off;
	const char *a_addrany;
	const char *a_loopback;
	int a_scoped;
} afdl [] = {
#define	N_INET6 0
#ifdef IPV6
	{PF_INET6, sizeof(struct in6_addr),
	 sizeof(struct sockaddr_in6),
	 offsetof(struct sockaddr_in6, sin6_addr),
	 in6_addrany, in6_loopback, 1},
#endif
#define	N_INET 1
	{PF_INET, sizeof(struct in_addr),
	 sizeof(struct sockaddr_in),
	 offsetof(struct sockaddr_in, sin_addr),
	 in_addrany, in_loopback, 0},
	{0, 0, 0, 0, NULL, NULL, 0},
};

struct explore {
	int e_af;
	int e_socktype;
	int e_protocol;
	const char *e_protostr;
	int e_wild;
#define WILD_AF(ex)		((ex)->e_wild & 0x01)
#define WILD_SOCKTYPE(ex)	((ex)->e_wild & 0x02)
#define WILD_PROTOCOL(ex)	((ex)->e_wild & 0x04)
};

static const struct explore explore[] = {
#ifdef IPV6
	{ PF_INET6, SOCK_DGRAM, IPPROTO_UDP, "udp", 0x07 },
	{ PF_INET6, SOCK_STREAM, IPPROTO_TCP, "tcp", 0x07 },
	{ PF_INET6, SOCK_RAW, ANY, NULL, 0x05 },
#endif
	{ PF_INET, SOCK_DGRAM, IPPROTO_UDP, "udp", 0x07 },
	{ PF_INET, SOCK_STREAM, IPPROTO_TCP, "tcp", 0x07 },
	{ PF_INET, SOCK_RAW, ANY, NULL, 0x05 },
	{ PF_UNSPEC, SOCK_DGRAM, IPPROTO_UDP, "udp", 0x07 },
	{ PF_UNSPEC, SOCK_STREAM, IPPROTO_TCP, "tcp", 0x07 },
	{ PF_UNSPEC, SOCK_RAW, ANY, NULL, 0x05 },
	{ -1, 0, 0, NULL, 0 },
};

#define PTON_MAX	16

static int str_isnumber(const char *);
static int explore_null(const struct addrinfo *,
	const char *, struct addrinfo **);
static int explore_numeric(const struct addrinfo *, const char *,
	const char *, struct addrinfo **);
static struct addrinfo *get_ai(const struct addrinfo *,
	const struct afd *, const char *);
static int get_portmatch(const struct addrinfo *, const char *);
static int get_port(struct addrinfo *, const char *, int);
static const struct afd *find_afd(int);

#define GET_AI(ai, afd, addr) \
do { \
	/* external reference: pai, error, and label free */ \
	(ai) = get_ai(pai, (afd), (addr)); \
	if ((ai) == NULL) { \
		error = EAI_MEMORY; \
		goto free; \
	} \
} while (/*CONSTCOND*/0)

#define GET_PORT(ai, serv) \
do { \
	/* external reference: error and label free */ \
	error = get_port((ai), (serv), 0); \
	if (error != 0) \
		goto free; \
} while (/*CONSTCOND*/0)

#define ERR(err) \
do { \
	/* external reference: error, and label bad */ \
	error = (err); \
	goto bad; \
	/*NOTREACHED*/ \
} while (/*CONSTCOND*/0)

#define MATCH_FAMILY(x, y, w) \
	((x) == (y) || (/*CONSTCOND*/(w) && ((x) == PF_UNSPEC || (y) == PF_UNSPEC)))
#define MATCH(x, y, w) \
	((x) == (y) || (/*CONSTCOND*/(w) && ((x) == ANY || (y) == ANY)))

void
irc_freeaddrinfo(struct addrinfo *ai)
{
  struct addrinfo *next;

  do
  {
    next = ai->ai_next;
    if (ai->ai_canonname)
      free(ai->ai_canonname);
    /* no need to free(ai->ai_addr) */
    free(ai);
    ai = next;
  }
  while(ai);
}

static int
str_isnumber(const char *p)
{
  char *ep;

  if (*p == '\0')
    return 0;
  ep = NULL;
  errno = 0;
  (void)strtoul(p, &ep, 10);
  if (errno == 0 && ep && *ep == '\0')
    return 1;
  else
    return 0;
}

int
irc_getaddrinfo(const char *hostname, const char *servname,
                const struct addrinfo *hints, struct addrinfo **res)
{
  struct addrinfo sentinel;
  struct addrinfo *cur;
  int error = 0;
  struct addrinfo ai;
  struct addrinfo ai0;
  struct addrinfo *pai;
  const struct explore *ex;

  memset(&sentinel, 0, sizeof(sentinel));
  cur = &sentinel;
  pai = &ai;
  pai->ai_flags = 0;
  pai->ai_family = PF_UNSPEC;
  pai->ai_socktype = ANY;
  pai->ai_protocol = ANY;
  pai->ai_addrlen = 0;
  pai->ai_canonname = NULL;
  pai->ai_addr = NULL;
  pai->ai_next = NULL;

  if (hostname == NULL && servname == NULL)
    return EAI_NONAME;
        if (hints) {
                /* error check for hints */
                if (hints->ai_addrlen || hints->ai_canonname ||
                    hints->ai_addr || hints->ai_next)
                        ERR(EAI_BADHINTS);
                if (hints->ai_flags & ~AI_MASK)
                        ERR(EAI_BADFLAGS);
                switch (hints->ai_family) {
                case PF_UNSPEC:
                case PF_INET:
#ifdef IPV6
                case PF_INET6:
#endif
                        break;
                default:
                        ERR(EAI_FAMILY);
                }
                memcpy(pai, hints, sizeof(*pai));

                /*
                 * if both socktype/protocol are specified, check if they
                 * are meaningful combination.
                 */
                if (pai->ai_socktype != ANY && pai->ai_protocol != ANY) {
                        for (ex = explore; ex->e_af >= 0; ex++) {
                                if (pai->ai_family != ex->e_af)
                                        continue;
                                if (ex->e_socktype == ANY)
                                        continue;
                                if (ex->e_protocol == ANY)
                                        continue;
                                if (pai->ai_socktype == ex->e_socktype && pai->ai_protocol != ex->e_protocol) {
                                        ERR(EAI_BADHINTS);
                                }
                        }
                }
        }

        /*
         * check for special cases.  (1) numeric servname is disallowed if
         * socktype/protocol are left unspecified. (2) servname is disallowed
         * for raw and other inet{,6} sockets.
         */
        if (MATCH_FAMILY(pai->ai_family, PF_INET, 1)
#ifdef IPV6
            || MATCH_FAMILY(pai->ai_family, PF_INET6, 1)
#endif
            ) {
                ai0 = *pai;     /* backup *pai */

                if (pai->ai_family == PF_UNSPEC) {
#ifdef IPV6
                        pai->ai_family = PF_INET6;
#else
                        pai->ai_family = PF_INET;
#endif
                }
                error = get_portmatch(pai, servname);
                if (error)
                        ERR(error);

                *pai = ai0;
        }

        ai0 = *pai;

        /* NULL hostname, or numeric hostname */
        for (ex = explore; ex->e_af >= 0; ex++) {
                *pai = ai0;

                /* PF_UNSPEC entries are prepared for DNS queries only */
                if (ex->e_af == PF_UNSPEC)
                        continue;

                if (!MATCH_FAMILY(pai->ai_family, ex->e_af, WILD_AF(ex)))
                        continue;
                if (!MATCH(pai->ai_socktype, ex->e_socktype, WILD_SOCKTYPE(ex)))
                        continue;
                if (!MATCH(pai->ai_protocol, ex->e_protocol, WILD_PROTOCOL(ex)))
                        continue;

                if (pai->ai_family == PF_UNSPEC)
                        pai->ai_family = ex->e_af;
                if (pai->ai_socktype == ANY && ex->e_socktype != ANY)
                        pai->ai_socktype = ex->e_socktype;
                if (pai->ai_protocol == ANY && ex->e_protocol != ANY)
                        pai->ai_protocol = ex->e_protocol;

                if (hostname == NULL)
                        error = explore_null(pai, servname, &cur->ai_next);
                else
                        error = explore_numeric(pai, hostname, servname, &cur->ai_next);

                if (error)
                        goto free;

                while (cur && cur->ai_next)
                        cur = cur->ai_next;
        }

        if (sentinel.ai_next)
                goto good;

        if (pai->ai_flags & AI_NUMERICHOST)
                ERR(EAI_NONAME);
        if (hostname == NULL)
                ERR(EAI_NODATA);

        if (sentinel.ai_next)
                error = 0;

        if (error)
                goto free;
        if (error == 0) {
                if (sentinel.ai_next) {
 good:
                        *res = sentinel.ai_next;
                        return SUCCESS;
                } else
                        error = EAI_FAIL;
        }
 free:
 bad:
        if (sentinel.ai_next)
                irc_freeaddrinfo(sentinel.ai_next);
        *res = NULL;
        return error;
}

/* hostname == NULL.
 * passive socket -> anyaddr (0.0.0.0 or ::)
 * non-passive socket -> localhost (127.0.0.1 or ::1)
 */
static int
explore_null(const struct addrinfo *pai, const char *servname, struct addrinfo **res)
{
  int s;
  const struct afd *afd;
  struct addrinfo *cur;
  struct addrinfo sentinel;
  int error;

	*res = NULL;
	sentinel.ai_next = NULL;
	cur = &sentinel;

	s = socket(pai->ai_family, SOCK_DGRAM, 0);
	if (s < 0) {
		if (errno != EMFILE)
			return 0;
	} else
		close(s);

	/*
	 * if the servname does not match socktype/protocol, ignore it.
	 */
	if (get_portmatch(pai, servname) != 0)
		return 0;

	afd = find_afd(pai->ai_family);
	if (afd == NULL)
		return 0;

	if (pai->ai_flags & AI_PASSIVE) {
		GET_AI(cur->ai_next, afd, afd->a_addrany);
		GET_PORT(cur->ai_next, servname);
	} else {
		GET_AI(cur->ai_next, afd, afd->a_loopback);
		GET_PORT(cur->ai_next, servname);
	}
	cur = cur->ai_next;

	*res = sentinel.ai_next;
	return 0;

free:
	if (sentinel.ai_next)
		irc_freeaddrinfo(sentinel.ai_next);
	return error;
}

static int
explore_numeric(const struct addrinfo *pai, const char *hostname,
	        const char *servname, struct addrinfo **res)
{
  const struct afd *afd;
  struct addrinfo *cur;
  struct addrinfo sentinel;
  int error;
  char pton[PTON_MAX];

  *res = NULL;
  sentinel.ai_next = NULL;
  cur = &sentinel;

	/* if the servname does not match socktype/protocol, ignore it. */
	if (get_portmatch(pai, servname) != 0)
		return 0;

	afd = find_afd(pai->ai_family);
	if (afd == NULL)
		return 0;

	switch (afd->a_af) {
#if 1 /*X/Open spec*/
	case AF_INET:
		if (inet_aton(hostname, (struct in_addr *)pton) == 1) {
			if (pai->ai_family == afd->a_af ||
			    pai->ai_family == PF_UNSPEC /*?*/) {
				GET_AI(cur->ai_next, afd, pton);
				GET_PORT(cur->ai_next, servname);
				while (cur && cur->ai_next)
					cur = cur->ai_next;
			} else
				ERR(EAI_FAMILY);
		}
		break;
#endif
	default:
		if (inet_pton(afd->a_af, hostname, pton) == 1) {
			if (pai->ai_family == afd->a_af ||
			    pai->ai_family == PF_UNSPEC /*?*/) {
				GET_AI(cur->ai_next, afd, pton);
				GET_PORT(cur->ai_next, servname);
				while (cur && cur->ai_next)
					cur = cur->ai_next;
			} else
				ERR(EAI_FAMILY);
		}
		break;
	}

	*res = sentinel.ai_next;
	return 0;

free:
bad:
	if (sentinel.ai_next)
		irc_freeaddrinfo(sentinel.ai_next);
	return error;
}

static struct addrinfo *
get_ai(const struct addrinfo *pai, const struct afd *afd, const char *addr)
{
  char *p;
  struct addrinfo *ai;
	
    ai = (struct addrinfo *)malloc(sizeof(struct addrinfo)
		+ (afd->a_socklen));
	if (ai == NULL)
		return NULL;

	memcpy(ai, pai, sizeof(struct addrinfo));
	ai->ai_addr = (struct sockaddr *)(void *)(ai + 1);
	memset(ai->ai_addr, 0, (size_t)afd->a_socklen);
	ai->ai_addrlen = afd->a_socklen;
	ai->ai_addr->sa_family = ai->ai_family = afd->a_af;
	p = (char *)(void *)(ai->ai_addr);
	memcpy(p + afd->a_off, addr, (size_t)afd->a_addrlen);
	return ai;
}

static int
get_portmatch(const struct addrinfo *ai, const char *servname)
{
  /* get_port does not touch first argument. when matchonly == 1.
   * LINTED const cast */
  return(get_port((struct addrinfo *)ai, servname, 1));
}

static int
get_port(struct addrinfo *ai, const char *servname, int matchonly)
{
  const char *proto;
  struct servent *sp;
  int port;
  int allownumeric;

  if (servname == NULL)
    return 0;
        switch (ai->ai_family) {
        case AF_INET:
#ifdef AF_INET6
        case AF_INET6:
#endif
		break;
	default:
		return 0;
	}

	switch (ai->ai_socktype) {
	case SOCK_RAW:
		return EAI_SERVICE;
	case SOCK_DGRAM:
	case SOCK_STREAM:
		allownumeric = 1;
		break;
	case ANY:
		allownumeric = 0;
		break;
	default:
		return EAI_SOCKTYPE;
	}

	if (str_isnumber(servname)) {
		if (!allownumeric)
			return EAI_SERVICE;
		port = atoi(servname);
		if (port < 0 || port > 65535)
			return EAI_SERVICE;
		port = htons(port);
	} else {
		switch (ai->ai_socktype) {
		case SOCK_DGRAM:
			proto = "udp";
			break;
		case SOCK_STREAM:
			proto = "tcp";
			break;
		default:
			proto = NULL;
			break;
		}

		if ((sp = getservbyname(servname, proto)) == NULL)
			return EAI_SERVICE;
		port = sp->s_port;
	}

	if (!matchonly) {
		switch (ai->ai_family) {
		case AF_INET:
			((struct sockaddr_in *)(void *)
			    ai->ai_addr)->sin_port = port;
			break;
#ifdef IPV6
		case AF_INET6:
			((struct sockaddr_in6 *)(void *)
			    ai->ai_addr)->sin6_port = port;
			break;
#endif
		}
	}

	return 0;
}

static const struct afd *
find_afd(int af)
{
  const struct afd *afd;

  if (af == PF_UNSPEC)
    return(NULL);

  for (afd = afdl; afd->a_af; afd++)
  {
    if (afd->a_af == af)
      return(afd);
  }

  return(NULL);
}
