/* motd.c: Message of the day.
 * Copyright (C) 2005 by MusIRCd Development.
 * $Id: motd.c,v 1.27 2005/07/13 03:38:43 musirc Exp $
 */

#include "motd.h"
#include "ircd.h"
#include "config.h"
#include "send.h"
#include "numeric.h"
#include "istring.h"
#include "sprintf.h"
#include "server.h"

void
init_message_file(MotdType motdType, const char *fileName, MessageFile *motd)
{
  strlcpy(motd->fileName, fileName, sizeof(motd->fileName));
  motd->motdType = motdType;
  motd->contentsOfFile = NULL;
}

/* This function split off so a server notice could be generated on a
 * user requested motd, but not on each connecting client. 
 */
int
send_message_file(struct Client *source_p, MessageFile *motdToPrint)
{
  MessageFileLine *linePointer;
  MotdType motdType;

  if (motdToPrint == NULL)
   return(-1);

  motdType = motdToPrint->motdType;

  switch (motdType)
  {
    case USER_MOTD:
      if (motdToPrint->contentsOfFile == NULL)
        sendto_one(source_p, form_str(ERR_NOMOTD), me.name, source_p->name);
      else
      {
        sendto_one(source_p, form_str(RPL_MOTDSTART),
                   me.name, source_p->name, me.name);

        for (linePointer = motdToPrint->contentsOfFile; linePointer;
             linePointer = linePointer->next)
          sendto_one(source_p, form_str(RPL_MOTD),
                     me.name, source_p->name, linePointer->line);

        sendto_one(source_p, form_str(RPL_ENDOFMOTD), me.name, source_p->name);
      }
      break;

    case USER_LINKS:
      if (motdToPrint->contentsOfFile != NULL)
      {
        for (linePointer = motdToPrint->contentsOfFile; linePointer;
             linePointer = linePointer->next)
          sendto_one(source_p, ":%s 364 %s %s",
                     me.name, source_p->name, linePointer->line);
      }
      break;

    case OPER_MOTD:
      if (motdToPrint->contentsOfFile != NULL)
      {
        sendto_one(source_p, form_str(RPL_OMOTDSTART),
                   me.name, source_p->name, me.name);
        sendto_one(source_p, form_str(RPL_OMOTD),
                   me.name, source_p->name, me.name);

        for (linePointer = motdToPrint->contentsOfFile; linePointer;
	     linePointer = linePointer->next)
          sendto_one(source_p, form_str(RPL_OMOTD),
                     me.name, source_p->name, linePointer->line);

        sendto_one(source_p, form_str(RPL_ENDOFOMOTD),
                   me.name, source_p->name);
      }
      break;
  case ISSUPPORT:
      if (motdToPrint->contentsOfFile != NULL)
      {
        for (linePointer = motdToPrint->contentsOfFile; linePointer;
             linePointer = linePointer->next)
        {
          sendto_one(source_p, form_str(RPL_ISUPPORT),
                     me.name, source_p->name, linePointer->line);
        }
      }
    break;
    default:
      break;
  }
  return(0);
}

/* inputs - pointer to MessageFileptr */
int 
read_message_file(MessageFile *MessageFileptr)
{
  struct stat sb;
  MessageFileLine *mptr = 0, *next_mptr = 0, *newMessageLine = 0, *currentMessageLine = 0;
  char buffer[MESSAGELINELEN], *p;
  FBFILE *file;

  for (mptr = MessageFileptr->contentsOfFile; mptr; mptr = next_mptr)
  {
    next_mptr = mptr->next;
    MyFree(mptr);
  }

  MessageFileptr->contentsOfFile = NULL;
  if (stat(MessageFileptr->fileName, &sb) < 0)
    return(-1);

  if ((file = fbopen(MessageFileptr->fileName, "r")) == NULL)
    return(-1);

  while (fbgets(buffer, sizeof(buffer), file))
  {
    if ((p = strchr(buffer, '\n')) != NULL)
      *p = '\0';
    newMessageLine = (MessageFileLine *)MyMalloc(sizeof(MessageFileLine));

    strlcpy(newMessageLine->line, buffer, sizeof(newMessageLine->line));
    newMessageLine->next = NULL;
 
    if (MessageFileptr->contentsOfFile != NULL)
    {
      if (currentMessageLine)
        currentMessageLine->next = newMessageLine;

      currentMessageLine = newMessageLine;
    }
    else
    {
      MessageFileptr->contentsOfFile = newMessageLine;
      currentMessageLine = newMessageLine;
    }
  }

  fbclose(file);
  return(0);
}

/* output       - pointer to new MessageFile
 * side effects - Use this when an internal Message File is wanted
 *                without reading an actual file. The MessageFile
 *                is init'ed, but must have content added to it through
 *                addto_MessageLine()
 */
MessageFile *
init_MessageLine(void)
{
  MessageFile *mf;
  MessageFileLine *mptr = NULL;

  mf = MyMalloc(sizeof(MessageFile));
  mf->motdType = ISSUPPORT;     /* XXX maybe pass it alone in args? */
  mptr = MyMalloc(sizeof(MessageFileLine));
  mf->contentsOfFile = mptr;
  return(mf);
}

/* inputs       - Pointer to existing MessageFile
 *              - New string to add to this MessageFile
 * side effects - Use this when an internal MessageFile is wanted
 *                without reading an actual file. Content is added
 *                to this MessageFile through this function.
 */
void
addto_MessageLine(MessageFile *mf, const char *str)
{
  MessageFileLine *mptr = mf->contentsOfFile, *nmptr = NULL;

  if (mptr == NULL)
  {
    mptr = MyMalloc(sizeof(MessageFileLine));
    strcpy(mptr->line, str);
    mf->contentsOfFile = mptr;
  }
  else
  {
    while (mptr->next != NULL)
      mptr = mptr->next;
    nmptr = MyMalloc(sizeof(MessageFileLine));
    strcpy(nmptr->line, str);
    mptr->next = nmptr;
  }
}

/* inputs       - pointer to the MessageFile to destroy
 * side effects - All the MessageLines attached to the given mf
 *                Are freed then one MessageLine is recreated
 */
void
destroy_MessageLine(MessageFile *mf)
{
  MessageFileLine *mptr = mf->contentsOfFile, *nmptr = NULL;

  if (mptr == NULL)
    return;

  for (mptr = mf->contentsOfFile; mptr != NULL; mptr = nmptr)
  {
    nmptr = mptr->next;
    MyFree(mptr);
  }
  mf->contentsOfFile = NULL;
}
